import pathlib
import io
import mimetypes
import http
from .ziper import Ziper

class Iconer(Ziper):
    '''
    Same as `Ziper` but used to serve file icons used in `Filer` and `Ziper`.
    Defines methods to help other objects find icon urls for files of different
    types. The root of `Iconer` is a zip file which contains image files in its
    toplevel. The name of each image file should be in the form `<name>.<ext>`
    where `<ext>` can be any extension and <name> can be any of the following:
    
        * The string `directory`. It makes the image used as the icon for
          directories.
        * A file extension. It makes the image used for file of this extension.
        * General mimetype (such as audio, video, text, ...). It makes the image
          used for this mimetype if the file extension image does not exist.
        * The string `generic`. It makes the image used as a fallback icon if
          none of the above was found.
    
    Any file without extension in `self.root` is ignored.
    '''
    
    def __init__(self, root=None, theme='humanity'):
        '''
        Overrides `Ziper.__init__()`. Does not take `iconer` argument
        and always uses `self` as its `self.iconer`.
        
        args:
        
            * root (path-like object): Similar to `Ziper` constructor. If
              `None` is given the `Iconer` finds its root based on `theme`
              argument. Defaults to `None`.
            * theme (str): Ignored if `root` is not `None`. If `root` is `None`,
              looks for the root in
              `<module path>/asset/themes/<theme name>.zip` where
              `<module path>` is the path of cofan library and <theme name>
              is the value given in this argument.
            * prefix: The root url of the `Iconer` added to the `Paterner`
              class. For example, if the iconer root url is `foo_bar/`, the
              string `foo_bar` must be given in this argument and also added
              to the `Patterner` object that will relay requests to this
              `Iconer`. Defaults to `__icons__/`.
        '''
        if root is None:
            root = str(pathlib.Path(__file__).parent / 'asset' / 'themes' / '{}.zip'.format(theme))
        
        super().__init__(root, iconer=self)
        self.icon_index = self.get_icons()
    
    def handle_get(self, handler, url, query=None, full_url=''):
      ext = url.casefold()
      mime = mimetypes.guess_type('a.{}'.format(url))[0]
      if mime is not None:
        mime = mime.split('/')[0]
      
      if ext in self.icon_index:
        url = self.icon_index[ext]
      elif mime is not None and mime in self.icon_index:
        url = self.icon_index[mime]
      elif 'generic' in self.icon_index:
        url = self.icon_index['generic']
      else:
        return self.short_response(http.HTTPStatus.NOT_FOUND)
      
      return super().handle_get(handler, url, query, full_url)
    
    def get_icons(self):
        '''
        Used in `self.__init__()`. Looks at the toplevel content of `self.root`
        for any files and makes a dictionary for each file. The keys of the
        dictionary are file names without extensions and the values are the file
        names with extensions. Any files without extension are ignored. The
        dictionary is used to look for icons without opening the zip file.
        '''
        icon_index = {}
        dir_ls = self.get_file_list('')
        for c in dir_ls['files']:
            name_ext = c.rsplit('.', 1)
            if len(name_ext) > 1:
                icon_index[name_ext[0]] = c
        
        return icon_index
    
    def get_icon(self, name):
        '''
        This method is to be used in other content providers to get icons.
        Returns the url of an icon for `name`. The icon is constructed in the
        following way:
        
            * The extension is extracted from `name`. If `name` has no
              extension, the full value of `name` is taken.
            * If there is a file in `self.root` named as the extension of `name`
              extracted in the previous step, the url of this file is
              returned. For example, if `name` is `foo.mp4`, this method will
              look for a file named `mp4.<extension>` where `<extension>` may
              be any string. The extension is case insensitive so `foo.mp4`
              will be the same as `foo.MP4`.
            * If there is no file found in the previous step, the mimetype is
              guessed and the general type is taken (such as audio,
              video, etc...).
            * If there is a file in `self.root` with the same name as the
              generel mimetype extracted in the previous step, the url of this
              file is returned. For example, if `name` is `foo.mp4`, this
              method will look for a file named `mp4.<anything>` first. If it
              there is no such file in `self.root`, the method will look for
              `video.<anything>`. `<anything>` may be any string.
            * If there is no file found in the previous step, this method looks
              for a file named `generic.<anything>` and the url of this file
              is returned.
            * If there is no file found in the previous step, an empty string
              is returned.
        '''
        ext = name.rsplit('.', 1)[-1].casefold()
        mime = mimetypes.guess_type(name)[0]
        if mime is not None:
            mime = mime.split('/')[0]
        if ext in self.icon_index:
            return (pathlib.Path(
                '/' + self.prefix) / self.icon_index[ext]).as_posix()
        elif mime is not None and mime in self.icon_index:
            return (pathlib.Path(
                '/' + self.prefix) / self.icon_index[mime]).as_posix()
        elif 'generic' in self.icon_index:
            return (pathlib.Path(
                '/' + self.prefix) / self.icon_index['generic']).as_posix()
        else:
            return ''




import logging
import re
import http
import traceback
from .base_provider import BaseProvider

logger = logging.getLogger(__name__)

class NoMatchFoundError(Exception):
  pass

class Patterner(BaseProvider):
    '''
    `BaseProvider` subclass that relays requests to other providers based on
    requested url pattern. The other providers are added to the `Patterner`
    instance with the request url pattern they should get. When the `Patterner`
    gets a request, it searches for a pattern that matches the beginning of the
    url. When found, the `Patterner` calls `handle_get()` method of the target
    provider, giving it the same parameters but with the prefix stripped from
    the beginning of the url.
    '''
    def __init__(self):
        self.patterns = []
        self.titles = []
    
    def add(self, pattern, provider, title=''):
        '''
        Adds a pattern and a provider to the `Patterner`.
        
        args:
        
            * pattern (str): A string containing the url prefix of the provider.
            * provider (`BaseProvider`): The provider to relay the request to.
        '''
        self.patterns.append((re.compile(pattern), provider))
        self.titles.append((pattern, title))
    
    def remove(self, pattern):
        '''
        Removes a pattern and its provider from the pattern list. If a pattern
        exists multiple times (which you should not do anyway), only the first
        occurance is removed.
        
        args:
            pattern (str): Pattern to remove.
        '''
        for c in range(len(self.patterns)):
            if self.patterns[c][0].pattern == pattern:
                self.patterns.pop(c)
        else:
            raise ValueError('pattern not found')
        
        for c in range(len(self.titles)):
            if self.titles[c][0] == pattern:
                self.patterns.pop(c)
    
    def find_provider(self, url):
        '''
        Searches the pattern list for a pattern that matches the beginning of
        the url and returns the corresponding provider and the pattern.
        '''
        for pattern, provider in self.patterns:
            if pattern.match(url):
                return provider, pattern
        else:
            raise NoMatchFoundError
    
    def handle_default(self, handler, url, *args, **kwargs):
      try:
        provider, pattern = self.find_provider(url)
        new_url = pattern.sub('', url, count=1)
        
        provider_method = provider.handle_method(handler)
        return provider_method(handler, new_url, *args, **kwargs)
      except NoMatchFoundError:
        return self.short_response(http.HTTPStatus.NOT_FOUND)
    
    def get_patterns(self):
        '''
        Returns the pattern strings added to the `Patterner`.
        '''
        patterns = []
        for c in self.patterns:
            patterns.append(c[0].pattern)
        return patterns
    
    def get_title(self, pattern):
        '''
        Returns the pattern title.
        
        args:
        
            * pattern (str): the pattern to look for its title.
        
        returns:
        
            The pattern title.
        '''
        
        for c in range(len(self.titles)):
            if self.titles[c][0] == pattern:
                return self.titles[c][1]




from .base_provider import BaseProvider
from .statuser import Statuser
from .patterner import Patterner
from .pattern_lister import PatternLister
from .ip_patterner import IPPatterner
from .filer import Filer
from .ziper import Ziper
from .iconer import Iconer
from .password_checker import PasswordChecker




import io
import http

class BaseProvider:
    '''
    basic provider class. each time a request is intended to this provider,
    the provider should return a response code, headers and body. it defines one
    method:
        
      * short_response(): helper method to send a response code and its
        description.
    
    You should override methods that respond to http requests you want to
    handle. A handler for an http request with method <method> has to be
    handle_<method>(). E.g. handle_get() for GET requests, handle_put() for
    PUT requests, and so on. A default handler is handle_default() which
    handles any request that does not have a handler defined.
    
    All handler methods take the same arguments and should return the same
    values. refer to handle_default() documentation below for arguments.
    
    You probably want to use one of `BaseProvider` subclasses or inherit it in
    your own class.
    '''
    
    @classmethod
    def handle_default(handler, url,
        query=None, body_data=None, full_url=''):
      '''
      Handles all incoming requests that do not have a defined handler method.
      Default implementation raises NotImplementedError.
      
      args:
      
        * handler (http.server.BaseHTTPRequestHandler): The object that called
          this method.
        * url (str): The url that was requested after removing all prefixes by
          other providers. Look at Patterner for information of how prefixes are
          stripped.
        * query (dict): Request query arguments. Defaults to empty dictionary.
          full_url (str): The full url that was requested without removing
          prefixes.
        * body_data: the request body. `BaseHandler` implementation
          supplies this argument as a `dict` only if the request is
          multipart/form-data. Otherwise, `BaseHandler` passes an empty dict.

        returns:
        
            * response (http.HTTPStatus): Response code.
            * headers (dict): Response headers.
            * content (binary file-like object): The response content.
            * postproc (callable): A callable that will be called after serving
              the content. This callable will be called as long as this handler
              method succeded regardless whether sending the content to the
              client succeded or not. The intention of this callable is to close
              all files other than content in case there are open files. For
              example, if content is a file inside a zip file, closing content
              is not enough without closing the parent zip file.
      '''
      
      raise NotImplementedError
    
    @staticmethod
    def short_response(response=http.HTTPStatus.OK, body=None):
        '''
        Convinience method which can be used to send a status code and its
        description. It returns the same values as `handle_get` but the code is
        specified as a parameter and the body defaults to a description of the
        code.
              
        args:
        
          * response (http.HTTPStatus): Response code to send. Defaults to
            `http.HTTPStatus.OK`.
          * body: The value to send in the body. It defaults to `None` which
            sends a short description of the code. If the body is not a bytes
            object, it is converted to a string and then encoded to utf8.
        
        returns:
        
          * response (http.HTTPStatus): Response code given in the args.
          * headers (dict): Response headers dict with `text/html` in
            `Content-Type` header and the length of the body in
            `Content-Length` header.
          * content (binary file-like object): Description of `response` as a
            utf-8 encoded string if `body` arg is `None`. Otherwise, returns
            `body` represented as utf-8 encoded string.
          * postproc: Always a value of `None`.
        '''
        
        if body is None:
            body = '{} {}'.format(response.value, response.name.replace('_', ' ')).encode('utf8')
        elif type(body) is not bytes:
            body = str(body).encode('utf8')
        
        f = io.BytesIO(body)
        length = f.seek(0, 2)
        f.seek(0)
        headers = {
                'Content-Type':'text/html',
                'Content-Length': length,
            }
        return response, headers, f, lambda: None
    
    def handle_method(self, handler):
      '''
      returns the handler for the incoming request based on the request method.
      e.g. returns `self.handle_get()` for GET request, `self.handle_put()`
      method for PUT request, and so on. if no handler for the request method is
      defined, returns `self.handle_default()`.
      '''
      method = f'handle_{handler.command.lower()}'
      if not hasattr(self, method):
        method = 'handle_default'
      return getattr(self, method)
      




from tempfile import TemporaryDirectory
from pathlib import Path
from threading import RLock
import shutil
import time
import logging
import traceback

logger = logging.getLogger(__name__)

def precheck_parent(f):
  def check(self, name, *args, **kwargs):
    if self.check_parent(name):
      return f(self, name, *args, **kwargs)
    else:
      raise ValueError(f'{name} not in root')
  
  return check

def precheck_max_files(f):
  def check(self, *args, **kwargs):
    if not self.check_max_files():
      return f(self, *args, **kwargs)
    else:
      raise ValueError('too many files')
  
  return check

def preclean(f):
  def clean(self, *args, **kwargs):
    print('cleaning')
    for c in self.path.iterdir():
      try:
        t = time.time()
        stat = c.stat()
        t = max(stat.st_atime, stat.st_mtime)
        print(c, time.time() - t, self.timeout)
        if time.time() - t >= self.timeout:
          c.unlink()
      except Exception as e:
        logging.warning('error while cleaning {c}')
        logging.warning(traceback.format_exc())
    
    return f(self, *args, **kwargs)
  
  return clean

class TFMan:
  '''
  Temporary file manager. Used in cofan by Filer to create a temporary file for
  each partially uploaded file until the upload completes. Then moves the
  temporary file to its upload path.
  '''
  def __init__(self, max_files=10, timeout=5*60):
    self.dir = TemporaryDirectory()
    self.path = Path(self.dir.name)
    self.max_files = max_files
    self.timeout = timeout
    self.lock = RLock()
  
  def check_parent(self, name):
    path = (self.path / name).resolve()
    return path.parent == self.path
  
  def check_max_files(self):
    return len(tuple(self.path.iterdir())) >= self.max_files
  
  @preclean
  @precheck_parent
  @precheck_max_files
  def create(self, name):
    path = (self.path / name).resolve()
    print('touching', path)
    path.touch(exist_ok=False)
      
  @precheck_parent
  def write(self, name, pos, data):
    path = (self.path / name).resolve()
    if path.is_file():
      if type(data) == bytes:
        mode = 'r+b'
      else:
        mode = 'r+'
      
      with path.open(mode) as f:
        written = 0
        f.seek(pos)
        while written < len(data):
          written += f.write(data[written:])
  
  @precheck_parent
  def append(self, name, data):
    path = (self.path / name).resolve()
    if type(data) == bytes:
      mode = 'ab'
    else:
      mode = 'a'
    
    with path.open(mode) as f:
      written = 0
      while written < len(data):
        written += f.write(data[written:])
  
  @precheck_parent
  def open(self, name, *args, **kwargs):
    path = (self.path / name).resolve()
    return path.open(*args, **kwargs)
  
  @precheck_parent
  def delete(self, name, missing_ok=False):
    path = self.path / name
    path.unlink(missing_ok=missing_ok)
  
  @precheck_parent
  def copy(self, name, dest):
    path = (self.path / name).resolve()
    shutil.copyfile(path, dest)
  
  def move(self, name, dest):
    path = (self.path / name).resolve()
    shutil.move(path, dest)
  
  @precheck_parent
  def size(self, name):
    return (self.path / name).stat().st_size




import http
from .base_provider import BaseProvider

class Statuser(BaseProvider):
    '''
    A subclass of `BaseProvider`. This provider is very similar to BaseProvider.
    The only difference is that it takes the response code in its constructor
    and sends this response code instead of OK.
    '''
    def __init__(self, response=http.HTTPStatus.NOT_FOUND):
      '''
      args:
      
        * response (http.HTTPStatus): The response code to send when
          `handle_default()` is called. It defaults to NOT_FOUND.
      '''
      self.response = response
    def handle_default(self, handler, url, query=None, body_data=None,
        full_url=''):
      return self.short_response(self.response)




'''
:Date: 2022-07-09
:Version: 1.0.2
:Authors:
    * Mohammad Alghafli <thebsom@gmail.com>

cofan is an http server library for serving files and any other things. You use
it to share content in the form of a website. The current classes give you the
following:
    
  * Serve the content of a local directory in a form similar to file browser
    with icons for directories and files based on their extension.
  * Serve the content of a local zip file the same way as the local
    directories.
  * Serve local html files as a web site.
  * Upload to existing directory.
  * Organize your urls in prefix trees.
  * Response differently for different ip addresses

It also supports requests of partial files to resume previously interrupted
download.

Here is an example of how to use it::
    
  from cofan import *
  
  #allows uploads
  from cofan.utils.tfman import TFMan
  import pathlib
  
  #our site will be like this:
  # /           (this is our root. will list all the branches below)
  # |
  # |- vid/
  # |  this branch is a file browser for videos
  # |
  # -- site/
  #    this will be a web site. just a collection of html files
  
  #lets make an http file browser and share our videos
  #first, we specify the icons used in the file browser
  #you can omit the theme. it defaults to `humanity`
  icons = Iconer(theme='humanity')
  
  #now we create a Filer and specify the path we want to serve
  vid = Filer(pathlib.Path.home() / 'Videos', iconer=icons, tfman=TFMan())
  
  #we also want to serve a web site. lets create another filer. since the root
  #directory of the site contains `index.html` file, the filer
  #will automatically redirect to it instead of showing a file browser
  #no file browser also means we do not need to specify `iconer`
  #parameter. you can still use it if you want but that would not be very
  #useful
  site = Filer(pathlib.Path.home() / 'mysite')
  
  #now we need to give prefixes to our branches
  #we create a patterner
  patterns = Patterner()
  
  #then we add the filers with their prefixes
  #make sure to add a trailing slash
  patterns.add('vid/', vid)
  patterns.add('site/', site)
  
  #now we have all branches. but what if the user types our root url?
  #the path we will get will be an empty string which is not a prefix of any
  #branch. that will be a 404
  #lets make the root list and other branches added to `patterns`
  #the branches will be shown like the file browser but now the icons will be
  #for the patterns instead of file extensions
  #we need to specify where the icons are taken from
  #the icons file should contain an icon named `vid.<ext>` and an icon named
  #`site.<ext>` where <ext> can be any extension.
  root = PatternLister(patterns, root=pathlib.Path.home() / 'icons.zip')
  
  #and we add our root to the patterns with empty prefix
  patterns.add('', root)
  
  #now we create our handler like in http.server. we give it our patterner
  handler = BaseHandler(patterns)
  
  #and create our server like in http.server
  srv = Server(('localhost', 8000), handler)
  
  #and serve forever
  srv.serve_forever()
  
  #now try to open your browser on http://localhost:8000/

This module can also be run as a main python script to serve files from a
directory.

commandline syntax::

    python -m cofan.py [-a <addr>] [-u] [-nu] [<root>]

options:

    * -a <addr>, --addr <addr>: specify the address and port to bind to. <addr>
      should be in the form `<ip>:<port>` where `<ip>` is the ip address and
      `<port>` is the tcp port. defaults to `localhost:8000`.
    * -u, --upload: allow uploads. defaults to False.
    * -nu, --no-upload: opposite -u option. disallow uploads. this is selected
      by default.

args:

    * root: The root directory to serve. Defaults to the current directory.
'''

import logging
import traceback
import re
import urllib.parse
import http
import http.server
import socketserver
from cgi import parse_header
import multipart
from .providers import *

try:
    from fileslice import Slicer
except ImportError:
    Slicer = None

__version__ = '1.0.0'

logger = logging.getLogger(__name__)

class BaseHandler(http.server.BaseHTTPRequestHandler):
    '''
    Base http handler class in cofan library. You give it a provider instance
    and gets the response of all requests from it. For example::
        
      myprovider = Filer('/path/to/my/directory')
      myhandler = BaseHandler(myprovider)
      srv = http.server.HTTPServer(('localhost', 80), myhandler)
        
    `BaseHandler` has 2 class attributes:
    
      * __header_modifiers__: A tuple of request headers which have modifier
        methods in this class. For each header present in a request,
        `self.mod_<header>()` is called where <header> is the header name. The
        modifier method must return the response, headers and content after
        any necessary modification. It must take the following arguments:
      
          * response: Last response status code after any previous
            modifications by other modifier mehtods.
          * headers: Last headers after any previous modifications by other
            modifier mehtods.
          * content: Last content file after any previous modifications by
            other modifier mehtods.
      
        This tuple currently has only one header: `Range`. See
        `self.mod_Range()` for the modification applied.
      * TIMEOUT: Connection timeout for each request connection.
    
    This class is probably good for most uses without creating a subclass. In
    case you want to subclass it, you may want to override `self.get_response()`
    or `self.default_response()`.
    '''
    
    
    TIMEOUT = 10
    __header_modifiers__ = ('Range',)
    
    def __init__(self, provider, max_size=2**30):
        '''
        args:
        
            provider: The provider to get content from.
            max_size: the maximum size for multipart/form-data body.
        '''
        
        self.provider = provider
        self.max_size = max_size
    
    def do_default(self):
      '''
      default request method handler. called when no do_* method is available
      for the method of the incoming request.
      '''
      
      handler = self.response_method
      
      self.request.settimeout(self.TIMEOUT)
      try:
        response, headers, content, postproc = handler()
      except NotImplementedError:
        response, headers, content, postproc = BaseProvider.short_response(
          http.HTTPStatus.NOT_IMPLEMENTED)
      except:
        traceback.print_exc()
        response, headers, content, postproc = BaseProvider.short_response(
          http.HTTPStatus.INTERNAL_SERVER_ERROR)
      
      try:
          self.send_response(response)
          for c in headers:
              self.send_header(c, headers[c])
          self.end_headers()
          data = content.read(1024)
          while data:
              self.wfile.write(data)
              data = content.read(1024)
          self.request.settimeout(None)
      finally:
          content.close()
          postproc()
    
    def do_GET(self):
        '''
        Calls `BaseProvider.handle_get()` from `self.provider` and sends the
        returned response, headers and content. After this process is done,
        the content file is closed and the returned post processing fucntion is
        called. This happens regardless of whether the process ended with
        success or not.
        '''
        
        self.request.settimeout(self.TIMEOUT)
        try:
          response, headers, content, postproc = self.get_response()
        except NotImplementedError:
          response, headers, content, postproc = BaseProvider.short_response(
            http.HTTPStatus.NOT_IMPLEMENTED)
        except:
          traceback.print_exc()
          response, headers, content, postproc = BaseProvider.short_response(
            http.HTTPStatus.INTERNAL_SERVER_ERROR)
        
        try:
            self.send_response(response)
            for c in headers:
                self.send_header(c, headers[c])
            self.end_headers()
            data = content.read(1024)
            while data:
                self.wfile.write(data)
                data = content.read(1024)
            self.request.settimeout(None)
        finally:
            content.close()
            postproc()
    
    def do_HEAD(self):
        '''
        Same as `self.do_GET()` but does not send any content.
        '''
        
        self.request.settimeout(self.TIMEOUT)
        try:
          response, headers, content, postproc = self.get_response()
        except NotImplementedError:
          response, headers, content, postproc = BaseProvider.short_response(
            http.HTTPStatus.NOT_IMPLEMENTED)
        except:
          traceback.print_exc()
          response, headers, content, postproc = BaseProvider.short_response(
            http.HTTPStatus.INTERNAL_SERVER_ERROR)
        
        try:
            self.send_response(response)
            for c in headers:
                self.send_header(c, headers[c])
            self.end_headers()
            self.request.settimeout(None)
        finally:
            content.close()
            postproc()
    
    def default_response(self):
        '''
        Same as `BaseProvider.get_response()` but calls
        `self.provider.handle_post()` instead of `self.provider.handle_get()`.
        It also does not modify headers (e.g. if there is Range request header,
        it does not respond with 204 but keep the response code and content
        without change.
        '''
        
        provider_method = self.provider.handle_method(self)
        
        body_data = None
        if self.headers['Content-Type'] is not None:
          content_type, pdict = parse_header(self.headers['Content-Type'])
          if content_type == 'multipart/form-data':
            body_data = self.parse_form()
        
        url = urllib.parse.urlparse(self.path)
        path = urllib.parse.unquote(url.path)[1:]
        query = urllib.parse.parse_qs(url.query)
        return provider_method(self, path, query, body_data, path)
    
    def parse_form(self):
      length = int(self.headers['Content-Length'])
      if length > self.max_size:
        raise ValueError(
          'body size ({}) is greater than max size ({})'.format(
            length, self.max_size
          )
        )
          
      body_data = {}
      
      def on_field(f):
        body_data.setdefault(
          f.field_name.decode('utf8'), []).append(f.value.decode('utf8'))
      def on_file(f):
        new_file = {}
        new_file['data'] = f.file_object
        new_file['name'] = f.file_name.decode('utf8')
        ext = new_file['name'].rsplit('.', 1)[-1]
        new_file['ext'] = ext
        new_file['data'].seek(0)
        body_data.setdefault(f.field_name.decode('utf8'), []).append(new_file)
      
      multipart.parse_form(self.headers, self.rfile, on_field, on_file)
      return body_data
    
    def get_response(self):
        '''
        Called by `self.do_GET()` and `self.do_HEAD()`. Parses the url  path and
        query string and replaces all url escapes. After that, it gets the
        response `self.provider.handle_get()`. Before returning the response,
        headers, content file and post processing callables, it calls any
        modifier method that should be called. If the method fails while
        applying modifications, it closes the content file and calls the post
        processing callable before propagating the exception.
        '''
        
        url = urllib.parse.urlparse(self.path)
        path = urllib.parse.unquote(url.path)[1:]
        query = urllib.parse.parse_qs(url.query)
        
        provider_method = self.provider.handle_method(self)
        
        response, headers, content, postproc = provider_method(
          self, path, query, full_url=path)
        
        try:
            response_group = int(str(response.value)[0])
            if response_group != 3 and response_group != 4 and response_group != 5:
                for c in self.__header_modifiers__:
                    if c in self.headers:
                        response, headers, content = getattr(self, 'mod_{}'.format(c))(response, headers, content)
        except:
            content.close()
            postproc()
            raise
        
        return response, headers, content, postproc
    
    def mod_Range(self, response, headers, content):
        '''
        Called when the `Range` header is present in the request. If the
        response status code is 200 (OK) and the content file is seekable, the
        status code is changed to 206 (Partial Content) and the content file is
        changed to a partial file pointing to the requested range.
        Response status code is changed to 416 (Requested Range Not Satisfiable)
        if the range start is not between 0 and total size or if the range end
        is not between start and total size.
        '''
        
        if (response == http.HTTPStatus.OK and content.seekable() and
                Slicer is not None):
            rng = self.headers['Range'].split('=')
            if rng[0] == 'bytes':
                rng = rng[1].split('-')
                if rng[0]:
                    start = int(rng[0])
                    if rng[1]:
                        end = int(rng[1])
                    else:
                        end = content.seek(0, 2) - 1
                else:
                    start = content.seek(0, 2) - int(rng[1])
                    end = content.seek(0, 2) - 1
            else:
                raise RuntimeError('invalid range')
            
            total_size = content.seek(0, 2)
            length = end - start + 1
            
            valid_start = 0 <= start < total_size
            valid_end = start <= end < total_size
            
            if valid_start and valid_end:
                rng = 'bytes {}-{}/{}'.format(start, end, total_size)
                
                headers['Content-Range'] = rng
                headers['Content-Length'] = length
                
                response = http.HTTPStatus.PARTIAL_CONTENT
            else:
                rng = 'bytes */{}'.format(total_size)
                
                headers['Content-Range'] = rng
                headers['Content-Length'] = length
                
                response = http.HTTPStatus.REQUESTED_RANGE_NOT_SATISFIABLE
            
            content = Slicer(content)(start, length)
        
        return response, headers, content
    
    def __getattr__(self, name):
      if name.startswith('do_'):
        return self.do_default
      else:
        getattr(super(), name)
    
    def __call__(self, *args, **kwargs):
        '''
        Creates and returns another instance of the same type as `self` with
        the same provider. Any arguments and keyword arguments are passed to the
        constructor of the base class of `self`.
        '''
        
        handler = type(self)(self.provider)
        super(type(handler), handler).__init__(*args, **kwargs)
        return handler
    
    @property
    def response_method(self):
      method = f'{self.command.lower()}_response'
      if not hasattr(self, method):
        method = 'default_response'
      return getattr(self, method)

class Server(socketserver.ThreadingMixIn, http.server.HTTPServer):
    '''
    same as `http.server.HTTPServer` but handles requests in daemon threads.
    '''
    
    daemon_threads = True




import http
import time
import hmac
import urllib.parse

from .base_provider import BaseProvider

class PasswordChecker(BaseProvider):
    '''
    Prevents unauthorized requests by checking password. Made mainly to be used
    with Filer uploads. Provides basic protection and should not be considered
    safe in any means.
    '''
    def __init__(self, provider, password, max_delay=60,
        safe_methods=('GET', 'HEAD')):
      self.provider = provider
      self.password = password.encode('utf-8')
      self.max_delay = max_delay
      self.safe_methods = safe_methods
      
      self.timestamps = {}
    
    def compute_hash(self, path, t):
      msg = urllib.parse.unquote_to_bytes(path) + t.encode('utf-8')
      hasher = hmac.new(self.password, msg, 'md5')
      return hasher.hexdigest()
    
    def add_timestamp(self, addr):
      #clear too old timestamps
      now = time.time()
      for c in list(self.timestamps):
        if now - self.timestamps[c] > self.max_delay:
          self.timestamps.pop(c)
      
      if addr not in self.timestamps:
        self.timestamps[addr] = now - self.max_delay
    
    def is_allowed(self, handler):
      if handler.command in self.safe_methods:
        return True
      
      if 'Verify-Hash-T' in handler.headers:
        t = handler.headers['Verify-Hash-T']
        ft = float(t)
      else:
        ft = -1
      
      addr = handler.client_address[0]
      self.add_timestamp(addr)
      
      if self.timestamps[addr] < ft:
        self.timestamps[addr] = ft
        remote_hash = handler.headers['Verify-Hash']
        local_hash = self.compute_hash(handler.path, t)
        return hmac.compare_digest(local_hash, remote_hash)
      else:
        return False
    
    def handle_default(self, handler, *args, **kwargs):
      if self.is_allowed(handler):
        provider_method = self.provider.handle_method(handler)
        return provider_method(handler, *args, **kwargs)
      else:
        return self.short_response(http.HTTPStatus.UNAUTHORIZED)




import pathlib
from .cofan import *
import re
import sys
import argparse
from .utils.misc import tree_patterner
from .utils.tfman import TFMan

parser = argparse.ArgumentParser(description='Serve files over http.')
parser.add_argument(
  '-a', '--addr', type=str, action='store', default='localhost:8000',
  help='http bind address in the form <ip>:<port>'
)
parser.add_argument(
  '-u', '--upload_password', action='store', default='',
  help='upload password. defaults to empty string which disables uploads'
)
parser.add_argument(
        'root', action='store', nargs='?', default='.',
        help='root directory to serve through http'
    )


args = parser.parse_args()

addr = args.addr.split(':')
addr[1] = int(addr[1])
root = args.root
upload_password = args.upload_password

main_tree = pathlib.Path(__file__).parent / 'main_tree'
env = {'root': root}
env['upload_password'] = upload_password
env['tfman'] = TFMan

localhost_checker = IPPatterner()
localhost_checker.add(re.compile(r'localhost|127.0.0.1'),
  tree_patterner(main_tree / 'localhost_tree', env))
localhost_checker.add(re.compile(r'.*'),
  tree_patterner(main_tree / 'others_tree', env))

handler = BaseHandler(localhost_checker)

srv = Server(tuple(addr), handler)
print('serving {} at {}:{}'.format(pathlib.Path(root).resolve(), *addr))
if addr[0] in ['', '0.0.0.0']:
    try:
        import netifaces
    except ImportError:
        pass
    else:
        ifaces = netifaces.interfaces()
        addrs = [netifaces.ifaddresses(c) for c in ifaces]
        family_addrs = []
        for c in addrs:
            if srv.address_family.value in c:
                family_addrs.extend(c[srv.address_family.value])
        final_addrs = [c['addr'] for c in family_addrs]
        
        print('serving at the following addresses:')
        for c in final_addrs:
            print(c)

try:
    srv.serve_forever()
except KeyboardInterrupt:
    pass




import re
import pathlib
import io
import http
from .base_provider import BaseProvider
from .filer import Filer

class PatternLister(Filer):
    '''
    Similar to `Filer` but instead of showing content of a directory, shows the
    prefixes added to a `Patterner`. It also provides icon urls for the
    prefixes. See `Patterner` for more details of the `Patterner` provider.
    '''
    def __init__(self, patterner, iconer=None, icons_prefix='__icons__/',
        default_format='html', exclude='__.*__', include='..*'):
      '''
      args:
      
          patterner: The `Patterner` to list its prefixes.
          iconer: Iconer which provides icons for the branches. Defaults to
          `None`.
          exclude: A regex `str` which will be matched to each prefix in
          `self.patterner`. If the prefix matches, it will not be listed.
          Defaults to `|__.*__/?` (empty prefix or any prefix that starts and
          ends with 2 underscores).
          include: A regex `str` which will be matched to each prefix in
          `self.patterner`. If the prefix does not matches, it will not be
          listed. Defaults to `.*` (any prefix).
          
      In order for a prefix to be listed and sent to the client, it must not
      match `exclude` and must match `include`. If any of them fails, the
      prefix is ignored.
      '''
      self.patterner = patterner
      self.iconer = iconer
      self.icons_prefix = icons_prefix
      self.page = type(self).PAGE
      self.default_format = default_format
      self.tfman = None
      
      self.exclude = re.compile(exclude)
      self.include = re.compile(include)
    
    def exists(self, url):
        '''
        overrides `Filer.exists()`.
        
        args:
        
            * url (path-like object): The recieved url to check.
        
        returns: True if `url` is an empty string
        '''
        return not url
    
    def is_dir(self, url):
        '''
        overrides `Filer.is_dir()`. same as `self.exists()`.
        
        args:
        
            * url (path-like object): The recieved url to check.
        
        returns: True if `url` is an empty string
        '''
        return self.exists(url)
    
    def is_file(self, url=''):
        '''
        overrides `Filer.is_file()`. always returns False
        
        args:
        
            * url (path-like object): The recieved url to check.
        
        returns: False
        '''
        return False
    
    def serve_dir_zip(self, handler, url, query, full_url):
      raise NotImplemented('Download as zip is not supported in PatternLister')
    
    def get_file_list(self, url):
        '''
        Overrides `Filer.get_file_list()` where the returned dict contains:
        
            * dirs: An empty list.
            * files: A list of patterns under `self`.
        
        args:
        
            * url (path-like object): The directory url to list its content.
        '''
        
        patterns = [
          {'name': c, 'title': self.patterner.get_title(c)}
          for c in self.patterner.get_patterns()
            if self.include.match(c) and not self.exclude.match(c)
        ]
        return {'dirs': [], 'files': patterns}
    



import logging
from pathlib import Path
import io
import os
import mimetypes
import http
import json
import traceback
import datetime
import zipfile
from base64 import b16encode
import threading

from .base_provider import BaseProvider

logger = logging.getLogger(__name__)

class Filer(BaseProvider):
    '''
    Lists directory contents and serves files from local file system based on
    the recieved url.
    
    Allows uploading files to this directory.
    
    For file uploads, consider using a TPChecker to prevent unauthorized
    uploads.
    '''
    
    HTTP_DATE_TEMPLATE = '%a, %d %b %Y %H:%M:%S GMT'
    MAX_AGE = 'max-age={}'.format(60)
    CHUNK_SIZE = 4 * 2**10
    PAGE = Path(__file__).parent / 'asset' / 'filer_main.html'
    
    def __init__(self, root, iconer=None, icons_prefix='__icons__/',
        default_format='html', tfman=None):
      '''
      args:
      
        * root (path-like object): Root directory or file to serve. Any
          requested url will be served starting from this directory.
        * iconer (Iconer): An instance of `Iconer` which will be used for the
          directories and files icons. Defaults to None which will display
          no icons.
        * icons_prefix (str): The prefix to be used for icons root url. Defaults
          to `__icons__/`.
        * default_format (str): can be the string `html` or `json`. This tells
          whether to serve the filer as an html page or as a json file.
        * tfman (TFMan): An instance of `cofan.utils.TFMan`. The temporary file
          manager to manage uploaded files until they are fully uploaded. If
          `None`, disables file uploads. Defaults to `None`.
      '''
      
      self.root = str(Path(root).resolve())
      self.iconer = iconer
      self.icons_prefix = icons_prefix
      self.page = type(self).PAGE
      self.default_format = default_format
      self.tfman = tfman
    
    def handle_get(self, handler, url, query=None, body_data=None,
        full_url=''):
      '''
        Serves files and directories. If `url` points to a file, the file is
        sent in the response. If `url` points to a directory, its content is
        sent as an html page if `query['format']` is the string `html` or as
        a JSON file if `query['format']` is the string `json`.
      '''
      if query is None:
        query = {}
      
      query.setdefault('format', [self.default_format])
      
      if url.startswith(self.icons_prefix):
        if self.iconer is not None:
          icon_url = url[len(self.icons_prefix):]
          return self.iconer.handle_get(handler, icon_url, query, full_url)
        else:
          return self.short_response(http.HTTPStatus.NOT_FOUND)
      elif self.is_dir(url):
        return self.serve_dir(handler, url, query, full_url)
      elif self.is_file(url):
        return self.serve_file(handler, url, query, full_url)
      else:
        return self.short_response(http.HTTPStatus.NOT_FOUND)
    
    def serve_dir(self, handler, url, query, full_url):
      '''
      Overrides `JFiler.serve_dir()`. Sends an html file containing
      subdirectories and files present in the requested directory in a similar
      way to file browsers. Icons are taken from `self.iconer`.
      '''
      
      path = Path(url)
      
      if self.is_file(path / 'index.html'):
        url = (path / 'index.html').as_posix()
        response = http.HTTPStatus.TEMPORARY_REDIRECT
        headers = {
                'Location': url,
                'Content-Length': 0
            }
        f = io.BytesIO()
        return response, headers, f, lambda: None
      
      if 'download' in query:
        return self.serve_dir_zip(handler, url, query, full_url)
      elif query['format'][0] == 'html':
        return self.serve_dir_html(handler, url, query, full_url)
      elif query['format'][0] == 'json':
        return self.serve_dir_json(handler, url, query, full_url)
      else:
        return short_response(http.HTTPStatus.BAD_REQUEST)
    
    def serve_dir_zip(self, handler, url, query, full_url):
      path = Path(self.root) / url
      rfd, wfd = os.pipe()
      rpipe, wpipe = open(rfd, 'rb'), open(wfd, 'wb')
      
      response = http.HTTPStatus.OK
      headers = {
        'Content-Type':'application/zip',
      }
      
      threading.Thread(target=self.zip_dir, args=(path, wpipe)).start()
      
      return response, headers, rpipe, lambda: None
    
    def serve_dir_html(self, handler, url, query, full_url):
      f = self.page.open('rb')
      
      response = http.HTTPStatus.OK
      length = f.seek(0, 2)
      f.seek(0)
      headers = {
        'Content-Type':'text/html',
        'Content-Length': length,
      }
      return response, headers, f, lambda: None
    
    def serve_dir_json(self, handler, url, query, full_url):
      dir_ls = self.get_file_list(url)
      dir_ls['base_url'] = full_url[:len(full_url)-len(url)]
      dir_ls['icons_prefix'] = self.icons_prefix
      
      if self.tfman is not None:
        dir_ls['upload-allowed'] = ''
      
      page = json.dumps(dir_ls)
      f = io.BytesIO(page.encode('utf8'))
      
      response = http.HTTPStatus.OK
      length = f.seek(0, 2)
      f.seek(0)
      headers = {
        'Content-Type':'application/json',
        'Content-Length': length,
      }
      
      return response, headers, f, lambda: None
    
    def serve_file(self, handler, url, query, full_url):
      path = Path(self.root) / url
      mtime = datetime.datetime.fromtimestamp(path.stat().st_mtime)
      ifmod = handler.headers['If-Modified-Since']
      if ifmod is not None:
        ifmod = datetime.datetime.strptime(ifmod, self.HTTP_DATE_TEMPLATE)
        if mtime <= ifmod:
          response = http.HTTPStatus.NOT_MODIFIED
          headers = {'Content-Length': 0}
          f = io.BytesIO(b'')
          return response, headers, f, lambda: None
      
      mime = mimetypes.guess_type(url)[0]
      f = path.open('br')
      
      response = http.HTTPStatus.OK
      length = f.seek(0, 2)
      f.seek(0)
      headers = {
        'Content-Length': length,
        'Last-Modified': mtime.strftime(self.HTTP_DATE_TEMPLATE),
        'Cache-Control': self.MAX_AGE,
      }
      if mime is not None:
        headers['Content-Type'] = mime
      
      return response, headers, f, lambda: None
    
    def handle_put(self, handler, url,
        query=None, body_data=None, full_url=''):
      
      if self.tfman is None:
        return self.short_response(http.HTTPStatus.BAD_REQUEST)
      
      tmp_name = b16encode(full_url.encode('utf8')).decode('ascii')
      
      if (
            handler.headers['Content-Length'] is None or
            int(handler.headers['Content-Length']) <= 0
          ):
        self.tfman.delete(tmp_name, True)
        
        response = http.HTTPStatus.OK
        f = io.BytesIO(b'')
        headers = {
          'Content-Length': 0,
        }
        
        return response, headers, f, lambda: None
      
      total_length = int(handler.headers['Content-Total-Length'])
      offset = int(handler.headers['Content-Offset'])
      length = int(handler.headers['Content-Length'])
      
      if offset == 0:
        self.tfman.create(tmp_name)
      
      with self.tfman.open(tmp_name, 'ab') as f:
        read_size = 0
        data = handler.rfile.read1(self.CHUNK_SIZE)
        read_size += len(data)
        f.write(data)
        while read_size < length:
          data = handler.rfile.read1(self.CHUNK_SIZE)
          read_size += len(data)
          f.write(data)
      
      if offset + length == total_length:
        self.save_file(tmp_name, url)
      
      response = http.HTTPStatus.OK
      f = io.BytesIO(b'')
      headers = {
        'Content-Length': 0,
      }
      
      return response, headers, f, lambda: None
    
    def save_file(self, fname, path):
      root_path = Path(self.root)
      new_file_path = root_path / path
      parent = new_file_path.parent.resolve()
      
      if parent.is_relative_to(root_path):
        self.tfman.move(fname, new_file_path)
      else:
        raise RuntimeError('tried to upload file outside my root dir')
    
    def exists(self, url):
        '''
        Returns `True` if the url points to an existing file or directory under
        `self.root`. Otherwise returns `False`.
        
        args:
        
            * url (path-like object): The recieved url to check.
        
        returns: True if `url` is an existing file or subdirectory in
            `self.root`. False otherwise.
        '''
        try:
            root = Path(self.root)
            path = root / url
            path = path.resolve()
            if path.samefile(self.root) or root in path.parents:
                return path.exists()
        except FileNotFoundError:
            pass
        
        return False
    
    def is_dir(self, url):
        '''
        Returns `True` if the url points to an existing directory under
        `self.root`. Otherwise returns `False`.
        
        args:
        
            * url (path-like object): The recieved url to check.
        
        return: True if `url` is an existing subdirectory of `self.root`. False
            otherwise.
        '''
        return self.exists(url) and (Path(self.root) / url).is_dir()
    
    def is_file(self, url):
        '''
        Returns `True` if the url points to an existing file under `self.root`.
        Otherwise returns `False`.
        
        args:
        
            * url (path-like object): The recieved url to check.
        
        returns: True if `url` is an existing file in `self.root` or one of its
            subdirectories. False otherwise.
        '''
        return self.exists(url) and (Path(self.root) / url).is_file()
    
    def get_file_list(self, url):
        '''
        Called by `self.handle_get`. Returns a dictionary containing 2
        members:
        
            * dirs: A list of directories under the directory pointed by url.
            * files: A list of files under the directory pointed by url.
        
        args:
        
            * url (path-like object): The directory url to list its content.
        '''
        if self.is_dir(url):
            path = Path(self.root) / url
            dir_ls = [x for x in path.glob('*') if not x.name.startswith('.')]
            
            dirs = sorted([x.name for x in dir_ls if x.is_dir()])
            files = sorted([x.name for x in dir_ls if x.is_file()])
            return {'dirs': dirs, 'files': files}
        else:
            raise ValueError('url is not pointing to a directory')
    
    def zip_dir(self, path, dest):
      try:
        with zipfile.ZipFile(dest, 'w') as f:
          for item in path.glob('**/*'):
            if item.is_file():
              f.write(item, str(item.relative_to(path)))
      finally:
        dest.close()




import pathlib
import urllib.parse
from pathlib import Path
import io
import mimetypes
import http
import zipfile
import traceback
import datetime
import json
from .filer import Filer

class Ziper(Filer):
    '''
    Same as `Filer` but serves the content of a zip file instead of
    a directory. The root in the constructor must be a zip file. Files served
    from this class are not seekable so resuming download is not possible for
    the content of a Ziper.
    '''
    
    def __init__(self, root, iconer=None, icons_prefix='__icons__/',
        default_format='html', tfman=None):
      super().__init__(root, iconer, icons_prefix, default_format, None)
    
    def handle_get(self, handler, url, query=None, body_data=None, full_url=''):
      '''
      Overrides `Filer.handle_get()`. Gives the same result as
      `Filer.handle_get()` but looks at the content of a zip file instead of
      a directory.
      '''
      
      if query is None:
        query = {}
      query.setdefault('format', [self.default_format])
      
      if url.startswith(self.icons_prefix):
        if self.iconer is not None:
          icon_url = url[len(self.icons_prefix):]
          return self.iconer.handle_get(handler, icon_url, query, full_url)
        else:
          return self.short_response(http.HTTPStatus.NOT_FOUND)
      
      with zipfile.ZipFile(self.root) as root:
        if self.is_dir(url, root):
          return self.serve_dir(handler, url, query, full_url, root)
        elif self.is_file(url, root):
          return self.serve_file(handler, url, query, full_url, root)
        else:
          return self.short_response(http.HTTPStatus.NOT_FOUND)
    
    def serve_dir(self, handler, url, query, full_url, root):
      '''
      Overrides `JFiler.serve_dir()`. Sends an html file containing
      subdirectories and files present in the requested directory in a similar
      way to file browsers. Icons are taken from `self.iconer`.
      '''
      
      path = Path(url)
      
      if self.is_file(path / 'index.html', root):
        url = (path / 'index.html').as_posix()
        response = http.HTTPStatus.TEMPORARY_REDIRECT
        headers = {
                'Location': url,
                'Content-Length': 0
            }
        f = io.BytesIO()
        return response, headers, f, lambda: None
      
      if 'download' in query:
        return self.serve_dir_zip(handler, url, query, full_url, root)
      elif query['format'][0] == 'html':
        return self.serve_dir_html(handler, url, query, full_url)
      elif query['format'][0] == 'json':
        return self.serve_dir_json(handler, url, query, full_url, root)
      else:
        return short_response(http.HTTPStatus.BAD_REQUEST)
    
    
    
    def serve_dir_json(self, handler, url, query, full_url, root):
      dir_ls = self.get_file_list(url, parent=bool(full_url), root=root)
      
      dir_ls['base_url'] = full_url[:len(full_url)-len(url)]
      dir_ls['icons_prefix'] = self.icons_prefix

      page = json.dumps(dir_ls)
      f = io.BytesIO(page.encode('utf8'))

      response = http.HTTPStatus.OK
      length = f.seek(0, 2)
      f.seek(0)
      headers = {
        'Content-Type':'application/json',
        'Content-Length': length,
      }

      return response, headers, f, lambda: None
    
    def serve_file(self, handler, url, query, full_url, root):
        '''
        Overrides `JFiler.serve_dir()`. Looks at the content of the zip file
        instead of a directory.
        
        This method has an additional optional argument:
        
            root(`zipfile.ZipFile`) Opened `self.root`. It is used in
            `self.handle_get()` to avoid opening and closing `self.root`
            multiple times. Defaults to None which means `self.root` will be
            opened and a new `zipfile.ZipFile` instance will be created.
        '''
        
        zinfo = root.getinfo(url).date_time
        mtime = datetime.datetime(*zinfo)
        ifmod = handler.headers['If-Modified-Since']
        if ifmod is not None:
            ifmod = datetime.datetime.strptime(ifmod, self.HTTP_DATE_TEMPLATE)
            if mtime <= ifmod:
                response = http.HTTPStatus.NOT_MODIFIED
                headers = {}
                f = io.BytesIO(b'')
                return response, headers, f, lambda: None
        
        mime = mimetypes.guess_type(url)[0]
        f = root.open(url)
        
        response = http.HTTPStatus.OK
        length = root.getinfo(url).file_size
        headers = {
                'Content-Length':length,
                'Last-Modified': mtime.strftime(self.HTTP_DATE_TEMPLATE),
                'Cache-Control': self.MAX_AGE
            }
        
        if mime is not None:
            headers['Content-Type'] = mime
        
        return response, headers, f, lambda: root.close()
    
    def get_file_list(self, url, parent=False, root=None):
        '''
        Overrides `Filer.get_file_list()`. Looks at the content of the zip
        file instead of a directory.
        '''
        
        if root is None:
          with zipfile.ZipFile(self.root) as root:
            return self.get_file_list(url, parent, root)
        
        dir_ls = [
          x for x in root.namelist() if
            x != url and
            x.startswith(url) and
            not x.split('/')[-1].startswith('.') and
            '/' not in x[0:-1].replace(url, '', 1)
        ]
        dirs = sorted([x.split('/')[-2] for x in dir_ls if x.endswith('/')])
        files = sorted([x.split('/')[-1] for x in dir_ls if not x.endswith('/')])
        
        return {'dirs': dirs, 'files': files}
    
    def exists(self, url, root):
        '''
        Overrides `Filer.exists()`. Looks at the content of the zip file
        instead of a directory.
        '''
        if not url:
            return pathlib.Path(self.root).exists()
        else:
            try:
                path = root.getinfo(url)
                return True
            except KeyError:
                return False
    
    def is_dir(self, url, root):
        '''
        Overrides `Filer.is_dir()`. Looks at the content of the zip file
        instead of a directory.
        '''
        if not url:
            return pathlib.Path(self.root).exists()
        else:
            try:
                path = root.getinfo(url).filename
                return not path or path.endswith('/')
            except KeyError:
                return False
        
    def is_file(self, url, root):
        '''
        Overrides `Filer.is_file()`. Looks at the content of the zip file
        instead of a directory.
        '''
        try:
            path = root.getinfo(url).filename
            return not path.endswith('/')
        except KeyError:
            return False




import pathlib
from .. import providers
from ..providers.patterner import Patterner
from importlib.util import spec_from_file_location, module_from_spec
import sys

def tree_patterner(path, env=None):
  '''
  creates url prefixes for the directory tree under path. each prefix is handled
  by the provider returned from the provider.py file in the directory.
  '''
  if env is None:
    env = {}
  
  path = pathlib.Path(path)
  patterner = Patterner()
  env.update({'self': patterner, 'providers': providers})
  
  for c in reversed(list(path.glob('**/provider.py'))):
    prefix = str(c.parent.relative_to(path)).strip('.') + '/'
    if prefix == '/':
      prefix = ''
    
    spec = spec_from_file_location('', c)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    provider = module.provider(env)
    if provider is not None:
      print('adding', prefix, provider)
      patterner.add(prefix, provider)
  
  return patterner




from pathlib import Path

def provider(env):
  if Path(env['root']).is_file():
    browser_cls = env['providers'].Ziper
    upload = False
  else:
    browser_cls = env['providers'].Filer
    upload = bool(env['upload_password'])
  
  if upload:
    browser = browser_cls(env['root'],
      iconer=env['providers'].Iconer(), tfman=env['tfman']())
    return env['providers'].PasswordChecker(browser, env['upload_password'])
  else:
    return env['providers'].Filer(env['root'], iconer=env['providers'].Iconer())
  




import http
import traceback
from .patterner import Patterner, NoMatchFoundError

class IPPatterner(Patterner):
    '''
    The same as `Patterner` but relays requests based on IP address patterns
    instead of url.
    '''
    
    def find_provider(self, addr):
        for pattern, provider in self.patterns:
            if pattern.match(addr):
                return provider, pattern
        else:
            raise NoMatchFoundError
    
    def handle_default(self, handler, *args, **kwargs):
      try:
        addr = handler.client_address[0]
        provider, pattern = self.find_provider(addr)
        
        provider_method = provider.handle_method(handler)
        return provider_method(handler, *args, **kwargs)
      except NoMatchFoundError:
        return self.short_response(http.HTTPStatus.UNAUTHORIZED)

